#!/usr/bin/perl -w

use strict;
use vars qw(@tree);

$| = 1;
init();
fetch();
optimize();
open(DB, ">ip2region.db-new");
print DB dumpdb();
close(DB);
rename("ip2region.db-new", "ip2region.db");
exit;

################################################################################
sub init {
  @tree = ("\x00\x00") x 256;
}
################################################################################
sub fetch {
  my ($first, $last, $name, $firstInt, $lastInt);

  open(FILE, "subnets.txt") or die $!;
  while (<FILE>) {
    chomp;
    $name = undef;
    if (m/^\s*(\d+\.\d+\.\d+\.\d+)\/(\d+)\s+([A-Z]{2}|\d+)\s*$/) {
      # 1.2.3.4./24 CODE format
      $first = $1;
      $name = $3;
#      $name = "0$name" if ($name < 10);
      $firstInt = unpack("N", pack('C4', split(/\./, $first)));
      $lastInt = $firstInt + (1 << (32 - $2)) - 1;
      $last = join('.', unpack('C4', pack('N', $lastInt)));
    }
    if (m/^\s*(\d+)\s+(\d+)\s+([A-Z]{2}|\d+)\s*$/) {
      # INT_START INT_LAST CODE format
      $firstInt = $1;
      $lastInt = $2;
      $name = $3;
      $first = join('.', unpack('C4', pack('N', $firstInt)));
      $last = join('.', unpack('C4', pack('N', $lastInt)));
    }
    if ($name =~ m/^\d+$/) {
      $name = int($name);
      next unless ($name);
      $name = pack('S', $name);
    }
    else {
      next unless ($name);
    }
    add($first, $last, $name) if (defined($name));
  }
  close(FILE);
}
################################################################################
sub add {
  my ($first, $last, $name) = @_;

  setRange(\@tree, $first, $last, $name);
}
################################################################################
sub setRange {
  my ($tree, $start, $end, $name) = @_;
  my (@start, @end);

  if (ref($tree) ne 'ARRAY') {
    if ($^W) {
      warn "\$tree is not an array object\n";
      warn "$tree: $start - $end: $name\n";
    }
    return;
  }
  @start = split(/\./, $start);
  @end = split(/\./, $end);
  if ($#start != $#end) {
    die "Length error: " . $#start . " and " . $#end . "\n";
    return;
  }
  if ($start[0] == $end[0]) {
    if (isequal(@start[1..$#start], ((0) x (@start - 1))) && isequal(@end[1..$#end], ((255) x (@end - 1)))) {
      # like xxx.0.0.0 ... xxx.255.255.255
      $$tree[$start[0]] = $name;
    }
    else {
      setRange(getSubTree($tree, $start[0]), join('.', @start[1..$#start]), join('.', @end[1..$#end]), $name);
    }
  }
  elsif (@start == 1) {
    for (my $index = $start[0]; $index <= $end[0]; $index++) {
      $$tree[$index] = $name;
    }
  }
  else {
    if (isequal(@start[1..$#start], ((0) x (@start - 1)))) {
      fillTree($tree, $start[0], $name);
    }
    else {
      setRange(getSubTree($tree, $start[0]), join('.', @start[1..$#start]), join('.', ('255') x (@start - 1)), $name);
    }
    for (my $index = $start[0] + 1; $index < $end[0]; $index++) {
      fillTree($tree, $index, $name);
    }
    if (isequal(@end[1..$#end], ((255) x (@end - 1)))) {
      fillTree($tree, $end[0], $name);
    }
    else {
      setRange(getSubTree($tree, $end[0]), join('.', ('0') x (@end - 1)), join('.', @end[1..$#end]), $name);
    }
  }
}
################################################################################
sub fillTree {
  my ($tree, $index, $name) = @_;
  my $i;

  if (ref($$tree[$index]) eq 'ARRAY') {
    $i = 0;
    foreach my $subtree (@{$$tree[$index]}) {
      if (ref($subtree) eq 'ARRAY') {
        fillTree($subtree, $i++, $name);
      }
      else {
        $subtree = $name if ($subtree eq "\x00\x00");
        print unpack("C", $subtree) . " !!! " . unpack("C", $name) . "\n" if ($subtree ne $name);
      }
    }
  }
  elsif ($$tree[$index] eq "\x00\x00") {
    $$tree[$index] = $name;
  }
  else {
    $$tree[$index] = $name;
    print unpack("C", $$tree[$index]) . " ??? " . unpack("C", $name) . "\n" if ($$tree[$index] ne $name);
  }
}
################################################################################
sub getSubTree {
  my ($tree, $index) = @_;
  my @a;

  if ($$tree[$index] eq "\x00\x00") {
    @a = ("\x00\x00") x 256;
    $$tree[$index] = \@a;
  }
  elsif ($$tree[$index] && !ref($$tree[$index])) {
    @a = ($$tree[$index]) x 256;
    $$tree[$index] = \@a;
  }
  return $$tree[$index];
}
################################################################################
sub isequal {
  my @arr = @_;
  my $len = scalar @arr;
  my $i;
  
  unless ($len % 2) {
    $len /= 2;
    for ($i = 0; $i < $len; $i++) {
      return 0 if ($arr[$i] ne $arr[$i + $len]);
    }
    return 1;
  }
  return 0;
}
################################################################################
sub optimize {
  my ($tree) = shift;
  my ($node);
  
  $tree = \@tree unless ($tree);
  for (my $index = 0; $index < 256; $index++) {
    $node = $$tree[$index];
    if (ref($node) eq 'ARRAY') {
      die scalar @$node unless (scalar @$node == 256);
      if (isequal(@$node, ($$node[0]) x 256)) {
        $$tree[$index] = $$node[0];
      }
      else {
        optimize($node);
      }
    }
  }
}
################################################################################
sub dumpdb {
  my ($offset, $tree) = @_;
  my (@offsets, @subtrees, $end, $result);

  $tree = \@tree unless ($tree);
  $offset = 0 unless (defined($offset));

  die unless (ref($tree) eq 'ARRAY');
  $end = $offset + 256 * 4;
  foreach my $node (@$tree) {
    if (ref($node) ne 'ARRAY') {
      push(@offsets, "\xFF\xFF" . $node);
    }
    else {
      push(@offsets, pack('N', $end));
      push(@subtrees, dumpdb($end, $node));
      $end += length($subtrees[-1]);
    }
  }
  die "Length error: " . length(join('', @offsets)) unless (length(join('', @offsets)) == 256*4);
  $result = join('', @offsets, @subtrees);
  die "Length error2" unless (length($result) == $end - $offset);
  return $result;
}
################################################################################

